#  _  __
# | |/ /___ ___ _ __  ___ _ _ ®
# | ' </ -_) -_) '_ \/ -_) '_|
# |_|\_\___\___| .__/\___|_|
#              |_|
#
# Keeper Commander
# Copyright 2022 Keeper Security Inc.
# Contact: ops@keepersecurity.com
#

import argparse
import base64
import collections
import datetime
import itertools
import json
import logging
import os
from typing import List, Optional, Any, Dict, Union, Sequence

from cryptography.hazmat.primitives import serialization
from cryptography.hazmat.primitives.asymmetric import ed25519
from urllib.parse import urlunparse

from .base import Command, RecordMixin, FolderMixin
from .helpers.timeout import parse_timeout
from .. import api, utils, vault, record_types, generator, crypto, attachment, record_facades, record_management
from ..breachwatch import BreachWatch
from ..commands import recordv3
from ..error import CommandError
from ..params import KeeperParams, LAST_RECORD_UID
from ..subfolder import try_resolve_path, find_folders, get_folder_path
from ..proto import APIRequest_pb2
from ..recordv3 import RecordV3


record_add_parser = argparse.ArgumentParser(prog='record-add', description='Add a record to folder.')
record_add_parser.add_argument('--syntax-help', dest='syntax_help', action='store_true',
                               help='Display help on field parameters.')
record_add_parser.add_argument('-f', '--force', dest='force', action='store_true', help='ignore warnings')
record_add_parser.add_argument('-t', '--title', dest='title', action='store', help='record title')
record_add_parser.add_argument('-rt', '--record-type', dest='record_type', action='store', help='record type')
record_add_parser.add_argument('-n', '--notes', dest='notes', action='store', help='record notes')
record_add_parser.add_argument('--folder', dest='folder', action='store',
                               help='folder name or UID to store record')
record_add_parser.add_argument('--self-destruct', dest='self_destruct', action='store',
                               metavar='<NUMBER>[(m)inutes|(h)ours|(d)ays]',
                               help='Time period record share URL is valid. The record will be deleted in your vault in 5 minutes since open')
record_add_parser.add_argument('fields', nargs='*', type=str,
                               help='load record type data from strings with dot notation')

record_update_parser = argparse.ArgumentParser(prog='record-update', description='Update a record.')
record_update_parser.add_argument('--syntax-help', dest='syntax_help', action='store_true',
                                  help='Display help on field parameters.')
record_update_parser.add_argument('-f', '--force', dest='force', action='store_true', help='ignore warnings')
record_update_parser.add_argument('-t', '--title', dest='title', action='store', help='modify record title')
record_update_parser.add_argument('-rt', '--record-type', dest='record_type', action='store', help='record type')
record_update_parser.add_argument('-n', '--notes', dest='notes', action='store', help='append/modify record notes')
record_update_parser.add_argument('-r', '--record', dest='record', action='store',
                                  help='record path or UID')
record_update_parser.add_argument('fields', nargs='*', type=str,
                                  help='load record type data from strings with dot notation')


append_parser = argparse.ArgumentParser(prog='append-notes', description='Append notes to an existing record.')
append_parser.add_argument('--notes', dest='notes', action='store', help='notes')
append_parser.add_argument('record', nargs='?', type=str, action='store', help='record path or UID')


delete_attachment_parser = argparse.ArgumentParser(
    prog='delete-attachment', description='Delete an attachment from a record.',
    usage="Example to remove two files for a record: delete-attachment {uid} --name secrets.txt --name photo.jpg")
delete_attachment_parser.add_argument('--name', dest='name', action='append', required=True, help='attachment file name or ID. Can be repeated.')
delete_attachment_parser.add_argument('record', action='store', help='record path or UID')


download_parser = argparse.ArgumentParser(prog='download-attachment', description='Download record attachments.')
download_parser.add_argument('-r', '--recursive', dest='recursive', action='store_true',
                             help='Download recursively through subfolders')
download_parser.add_argument('--out-dir', dest='out_dir', action='store', help='Local folder for downloaded files')
download_parser.add_argument('--preserve-dir', dest='preserve_dir', action='store_true',
                             help='Preserve vault folder structure')
download_parser.add_argument('--record-title', dest='record_title', action='store_true',
                             help='Add record title to attachment file.')
download_parser.add_argument('records', nargs='*', help='Record/Folder path or UID')


upload_parser = argparse.ArgumentParser(prog='upload-attachment', description='Upload record attachments.')
upload_parser.add_argument('--file', dest='file', action='append', required=True, help='file name to upload')
upload_parser.add_argument('record', action='store', help='record path or UID')


record_fields_description = '''
Commander supports two types of records:
1. Typed
2. Legacy (update only)

The content of Typed record is defined by schema. The schema name is stored on record "type" field
To view all available record types:  "record-type-info" or "rti"
To view fields for particular record type:  "record-type-info --list-record <record type>"  "rti -lt login"
To view field information type: "record-type-info --list-field <field type>"  "rti -lf host"

The Commander supports the following syntax for record fields:
[<FIELD_SET>][<FIELD_TYPE>][<FIELD_LABEL>]=[FIELD_VALUE]
Field components are separated with a dot (.)
1. FIELD_SET: Optional. 'f' or 'c'. Field section: field/f or custom/c
2. FIELD_TYPE: Mandatory for main fields optional for custom. if omitted 'text' field type is assumed
3. FIELD_LABEL: Optional. When adding multiple custom fields of the same type make sure the label is unique.
4. FIELD_VALUE: Optional. If is empty them field to be deleted. The field value content depends on field type.
Example:   "url.Web URL=https://google.com"

Field types are case sensitive
Field labels are case insensitive

Use full <FIELD_TYPE>.<FIELD_LABEL> syntax when field label collides with field type.
Example:  "password"          "password" field with no label
          "text.password"     "text" field with "password" label
          "Password"          "text" field with "Password" label

Use full <FIELD_TYPE>.<FIELD_LABEL> syntax when field label contains a dot (.)
Example:   "google.com"       Incorrect field type google
           "text.google.com"  Field type "text" field label "google.com"

If field label contains equal sign '=' then double it.
If field value starts with equal sign then prepend a value with space
Example:
    text.aaa==bbb=" =ccc"     sets custom field with label "aaa=bbb" to "=ccc"

The Legacy records define the following field types.
1. login
2. password
3. url
4. oneTimeCode

All records support:
3. Custom Fields: Any field that is not the pre-defined field is added to custom field section.
   "url.Web URL=https://google.com"
4. File Attachments:   "file=@<FILE_NAME>"

Supported record type field values:
Field Type        Description            Value Type     Examples
===========       ==================     =========+     =====================================
file              File attachment                       @file.txt
date              Unix epoch time.       integer        1668639533 | 2022-11-16T10:58:53Z | 2022-11-16
host              host name / port       object         {"hostName": "", "port": ""} 
                                                        192.168.1.2:4321
address           Address                object         {"street1": "", "street2": "", "city": "", "state": "",
                                                         "zip": "", "country": ""}
                                                        123 Main St, SmallTown, CA 12345, USA
phone             Phone                  object         {"region": "", "number": "", "ext": "", "type": ""}
                                                        Mobile: US (555)555-1234
name              Person name            object         {"first": "", "middle": "", "last": ""}
                                                        Doe, John Jr. | Jane Doe
securityQuestion  Security Q & A         array of       [{"question": "", "answer": ""}]
                                         objects        What city you were ...? city; What is the name of ...? name
paymentCard       Payment Card           object         {"cardNumber": "", "cardExpirationDate": "", "cardSecurityCode": ""}
                                                        4111111111111111 04/2026 123
bankAccount       Bank Account           object         {"accountType": "", "routingNumber": "", "accountNumber": ""}
                                                        Checking: 123456789 987654321
keyPair           Key Pair               object         {"publicKey": "", "privateKey": ""}

oneTimeCode       TOTP URL               string         otpauth://totp/Example?secret=JBSWY3DPEHPK3PXP&issuer=Keeper
note              Masked multiline text  string
multiline         Multiline text         string
secret            Masked text            string
login             Login                  string
email             Email                  string         'name@company.com'
password          Password               string
url               URL                    string         https://google.com/
text              Free form text         string         This field type generally has a label

$<ACTION>[:<PARAMS>, <PARAMS>]   executes an action that returns a field value

Value                   Field type         Description                      Example
====================    ===============    ===================              ==============
$GEN:[alg],[n]          password           Generates a random password      $GEN:dice,5
                                           Default algorith is rand         alg: [rand | dice | crypto]
                                           Optional: password length
$GEN                    oneTimeCode        Generates TOTP URL
$GEN:[alg,][enc]        keyPair            Generates a key pair and         $GEN:ec,enc
                                           optional passcode                alg: [rsa | ec | ed25519], enc
$JSON:<JSON TEXT>       any object         Sets a field value as JSON
                                           phone.Cell=$JSON:'{"number": "(555) 555-1234", "type": "Mobile"}'

PAM records require a PAM Configuration and additional commands for complete setup.
Example setup: A machine with user set up for rotation, connections, and tunneling:

mkdir gwapp -sf -a
secrets-manager app create gwapp1
secrets-manager share add --app=gwapp1 --secret=SHARED_FOLDER_UID --editable

pam gateway new --name=gateway1 --application=gwapp1 --config-init=b64 --return_value
pam config new --environment=local --title=config1 --gateway=gateway1 -sf=SHARED_FOLDER_UID \
    --connections=on --tunneling=on --rotation=on --remote-browser-isolation=on

record-add --folder=SHARED_FOLDER_UID --title=admin1 -rt=pamUser login=admin1 password="$GEN:rand,16"
record-add --folder=SHARED_FOLDER_UID --title=user1  -rt=pamUser login=user1  password="$GEN:rand,16"
record-add --folder=SHARED_FOLDER_UID --title=machine1 -rt=pamMachine \
  pamHostname="$JSON:{\"hostName\": \"127.0.0.1\", \"port\": \"22\"}"

pam tunnel edit PAM_MACHINE_UID --configuration=PAM_CONFIG_UID --enable-tunneling
pam connection edit PAM_MACHINE_UID --configuration=PAM_CONFIG_UID \
  --connections=on \
  --protocol=ssh \
  --admin-user=ADMIN_USER_UID

pam rotation edit --config=PAM_CONFIG_UID \
  --record=PAM_USER_UID \
  --resource=PAM_MACHINE_UID \
  --admin-user=ADMIN_USER_UID \
  --on-demand --enable --force
'''


ParsedFieldValue = collections.namedtuple('ParsedFieldValue', ['section', 'type', 'label', 'value'])


class RecordEditMixin:
    def __init__(self):
        self.warnings = []

    def on_warning(self, message):
        if message:
            self.warnings.append(message)

    def on_info(self, message):
        logging.info(message)

    @staticmethod
    def parse_field(field):   # type: (str) -> ParsedFieldValue
        if not isinstance(field, str):
            raise ValueError('Incorrect field value')

        name, sel, value = field.partition('=')
        if not sel:
            raise ValueError(f'Expected: <field>=<value>, got: {field}; Missing `=`')
        if not name:
            raise ValueError(f'Expected: <field>=<value>, got: {field}; Missing <field>')
        while value.startswith('='):
            name1, sel, value1 = value[1:].partition('=')
            if sel:
                name += sel + name1
                value = value1
            else:
                break

        field_section = ''
        if name.startswith('f.') or name.startswith('c.'):
            field_section = name[0]
            name = name[2:]
        if not name:
            raise ValueError(f'Expected: <field>=<value>, got: {field}; Missing field type or label')

        field_type, sep, field_label = name.partition('.')
        if not sep:
            if field_type in ('file'):
                pass
            elif field_type not in record_types.RecordFields:
                field_label = field_type
                field_type = ''
        return ParsedFieldValue(field_section, field_type, field_label, value.strip())

    def assign_legacy_fields(self, record, fields):
        # type: (vault.PasswordRecord, List[ParsedFieldValue]) -> None
        if not isinstance(record, vault.PasswordRecord):
            raise ValueError('Expected legacy record')
        if not isinstance(fields, list):
            raise ValueError('Fields parameter: expected array of strings')

        action_params = []
        for parsed_field in fields:
            if parsed_field.type == 'login':
                record.login = parsed_field.value
            elif parsed_field.type == 'password':
                if self.is_generate_value(parsed_field.value, action_params):
                    record.password = self.generate_password(action_params)
                else:
                    record.password = parsed_field.value
            elif parsed_field.type == 'url':
                record.link = parsed_field.value
            elif parsed_field.type == 'oneTimeCode':
                if self.is_generate_value(parsed_field.value, action_params):
                    record.totp = self.generate_totp_url()
                else:
                    record.totp = parsed_field.value
            else:
                field_type = parsed_field.type
                field_label = parsed_field.label
                if field_type and not field_label:
                    field_label = field_type
                index = next((i for i, x in enumerate(record.custom) if x.name.lower() == field_label.lower()), -1)
                if parsed_field.value:
                    if 0 <= index < len(record.custom):
                        record.custom[index].value = parsed_field.value
                    else:
                        record.custom.append(vault.CustomField.new_field(field_label, parsed_field.value))
                else:
                    if 0 <= index < len(record.custom):
                        record.custom.pop(index)

    def is_json_value(self, value, parameters):   # type: (str, List[Any]) -> Optional[bool]
        if value.startswith('$JSON'):
            value = value[5:]
            if value.startswith(':'):
                j_str = value[1:]
                if j_str and isinstance(parameters, list):
                    try:
                        parameters.append(json.loads(j_str))
                    except Exception as e:
                        self.on_warning(f'Invalid JSON value: {j_str}: {e}')
            return True

    @staticmethod
    def is_generate_value(value, parameters):    # type: (str, List[str]) -> Optional[bool]
        if value.startswith("$GEN"):
            value = value[4:]
            if value.startswith(':'):
                gen_parameters = value[1:]
                if gen_parameters and isinstance(parameters, list):
                    parameters.extend((x.strip() for x in gen_parameters.split(',')))
            return True

    @staticmethod
    def generate_key_pair(key_type, passphrase):  # type: (str, str) -> dict
        if key_type == 'ec':
            private_key, public_key = crypto.generate_ec_key()
        elif key_type == 'ed25519':
            private_key = ed25519.Ed25519PrivateKey.generate()
            public_key = private_key.public_key()
        else:
            private_key, public_key = crypto.generate_rsa_key()
        encryption = serialization.BestAvailableEncryption(passphrase.encode()) \
            if passphrase else serialization.NoEncryption()

        # noinspection PyTypeChecker
        pem_private_key = private_key.private_bytes(
            encoding=serialization.Encoding.PEM,
            format=serialization.PrivateFormat.PKCS8,
            encryption_algorithm=encryption)
        # noinspection PyTypeChecker
        pem_public_key = public_key.public_bytes(
            encoding=serialization.Encoding.PEM,
            format=serialization.PublicFormat.SubjectPublicKeyInfo)
        return {
            'privateKey': pem_private_key.decode(),
            'publicKey': pem_public_key.decode(),
        }

    @staticmethod
    def generate_password(parameters=None):   # type: (Optional[Sequence[str]]) -> str
        if isinstance(parameters, (tuple, list, set)):
            algorithm = next((x for x in parameters if x in ('rand', 'dice', 'crypto')), 'rand')
            length = next((x for x in parameters if x.isnumeric()), None)
            if isinstance(length, str) and len(length) > 0:
                try:
                    length = int(length)
                except ValueError:
                    pass
        else:
            algorithm = 'rand'
            length = None

        if algorithm == 'crypto':
            gen = generator.CryptoPassphraseGenerator()
        elif algorithm == 'dice':
            if isinstance(length, int):
                if length < 1:
                    length = 1
                elif length > 40:
                    length = 40
            else:
                length = 5
            gen = generator.DicewarePasswordGenerator(length)
        else:
            if isinstance(length, int):
                if length < 4:
                    length = 4
                elif length > 200:
                    length = 200
            else:
                length = 20
            gen = generator.KeeperPasswordGenerator(length=length)
        return gen.generate()

    @staticmethod
    def generate_totp_url():
        secret = base64.b32encode(crypto.get_random_bytes(20)).decode()

        return f'otpauth://totp/Commander?secret={secret}&issuer=Keeper'

    def validate_json_value(self, field_type, field_value):        # type: (str, Any) -> Any
        record_field = record_types.RecordFields.get(field_type)
        if not record_field:
            return field_value
        value_type = record_types.FieldTypes[record_field.type]
        if isinstance(value_type.value, dict):
            f_fields = set(value_type.value.keys())
            if isinstance(field_value, (list, dict)):
                if isinstance(field_value, list):
                    if record_field.multiple != record_types.Multiple.Always:
                        self.on_warning(f'Field \"{record_field.name}\" does not support multiple values')
                d_rs = []
                for dv in field_value if isinstance(field_value, list) else [field_value]:
                    if isinstance(dv, dict):
                        v_fields = set(dv.keys())
                        v_fields.difference_update(f_fields)
                        if len(v_fields) > 0:
                            self.on_warning(f'Field \"{record_field.name}\": '
                                            f'Properties \"{", ".join(v_fields)}\" are not supported.')
                        for key in f_fields:
                            if key not in dv:
                                dv[key] = ''
                        d_rs.append(dv)
                    else:
                        self.on_warning(f'Field \"{record_field.name}\": Incorrect value: \"{json.dumps(dv)}\"')
                        return
                if len(d_rs) > 1:
                    return d_rs
                elif len(d_rs) == 1:
                    return d_rs[0]
            else:
                self.on_warning(f'Field \"{record_field.name}\" ')
        elif isinstance(field_value, type(value_type.value)):
            return field_value
        else:
            self.on_warning(f'Field \"{record_field.name}\": Incorrect value: \"{field_value}\" ')

    @staticmethod
    def validate_notes(notes):    # type: (str) -> str
        if isinstance(notes, str):
            notes = notes.replace('\\\\n', '\x00')
            notes = notes.replace('\\n', '\n')
            notes = notes.replace('\x00', '\\n')
        return notes

    @staticmethod
    def adjust_typed_record_fields(record, typed_fields):    # type: (vault.TypedRecord, List[Dict]) -> Optional[bool]
        new_fields = []
        old_fields = [x for x in itertools.chain(record.fields, record.custom) if x.value]
        should_rebuild = False
        for typed_field in typed_fields:
            if not isinstance(typed_field, dict):
                return
            field_type = typed_field.get('$ref')
            if not field_type:
                return
            field_label = typed_field.get('label') or ''
            required = typed_field.get('required')
            rf = record_types.RecordFields.get(field_type)
            ignore_label = rf.multiple == record_types.Multiple.Never if rf else False

            # exact match
            field = next((x for x in old_fields if x.type == field_type and
                          (ignore_label or (x.label or '') == field_label)), None)
            # match first not empty
            if not field:
                if field_label:
                    field = next((x for x in old_fields if x.type == field_type and not x.label and x.value), None)
                else:
                    field = next((x for x in old_fields if x.type == field_type and x.value), None)

            if field:
                old_fields.remove(field)
                new_fields.append(field)
                field.required = required
                if field.label != field_label:
                    field.label = field_label
                    should_rebuild = True
                continue

            default_value = None
            if field_type == 'appFiller' and 'appFillerData' in typed_field:
                default_value = typed_field['appFillerData']
            field = vault.TypedField.new_field(field_type, default_value, field_label)
            field.required = required
            new_fields.append(field)
            should_rebuild = True

        custom = []
        if len(old_fields) > 0:
            custom.extend(old_fields)
            should_rebuild = True

        if should_rebuild:
            record.fields.clear()
            record.fields.extend(new_fields)
            record.custom.clear()
            record.custom.extend((x for x in custom if x.value))

        return should_rebuild

    def assign_typed_fields(self, record, fields):
        # type: (vault.TypedRecord, List[ParsedFieldValue]) -> None
        if not isinstance(record, vault.TypedRecord):
            raise ValueError('Expected typed record')
        if not isinstance(fields, list):
            raise ValueError('Fields parameter: expected array of fields')

        parsed_fields = collections.deque(fields)
        while len(parsed_fields) > 0:
            parsed_field = parsed_fields.popleft()
            field_type = parsed_field.type or 'text'
            field_label = parsed_field.label or ''
            skip_validation = not parsed_field.value or parsed_field.value.startswith('$JSON')
            if field_type not in record_types.RecordFields:
                if not skip_validation:
                    self.on_warning(f'Field type \"{field_type}\" is not supported. Field: {field_type}.{field_label}')
                    continue
            rf = record_types.RecordFields.get(field_type)
            ignore_label = rf.multiple == record_types.Multiple.Never if rf else False

            record_field = None    # type: Optional[vault.TypedField]
            is_field = False
            if parsed_field.section == 'f':   # ignore label
                fs = [x for x in record.fields if x.type == field_type and isinstance(x, vault.TypedField)]
                if len(fs) == 0:
                    self.on_warning(f'Field type \"{field_type}\" is not found for record type {record.record_type}')
                elif len(fs) == 1:
                    record_field = fs[0]
                else:
                    fs = [x for x in fs if (x.label or '').lower() == field_label.lower()]
                    if len(fs) == 0:
                        self.on_warning(
                            f'Field type \"{field_type}\" is not found for record type {record.record_type}')
                    else:
                        record_field = fs[0]
                is_field = True
            else:
                f_label = field_label.lower()
                record_field = next(
                    (x for x in record.fields
                     if (not parsed_field.type or x.type == parsed_field.type) and
                     (ignore_label or (x.label or '').lower() == f_label)), None)
                if record_field:
                    is_field = True
                else:
                    record_field = next(
                        (x for x in record.custom
                         if (not parsed_field.type or x.type == parsed_field.type) and
                         (ignore_label or (x.label or '').lower() == f_label)), None)
                    if record_field is None:
                        if not parsed_field.value:
                            continue
                        record_field = vault.TypedField.new_field(field_type or 'text', None, field_label)
                        record.custom.append(record_field)
            if not record_field:
                continue

            if parsed_field.value:
                value = None
                action_params = []
                if self.is_generate_value(parsed_field.value, action_params):
                    if record_field.type == 'password':
                        value = self.generate_password(action_params)
                    elif record_field.type in ('oneTimeCode', 'otp'):
                        value = self.generate_totp_url()
                    elif record_field.type in ('keyPair', 'privateKey'):
                        should_encrypt = 'enc' in action_params
                        passphrase = self.generate_password() if should_encrypt else None
                        key_type = next((x for x in action_params if x in ('rsa', 'ec', 'ed25519')), 'rsa')
                        value = self.generate_key_pair(key_type, passphrase)
                        if passphrase:
                            parsed_fields.append(ParsedFieldValue('', 'password', 'passphrase', passphrase))
                    else:
                        self.on_warning(f'Cannot generate a value for a \"{record_field.type}\" field.')
                elif self.is_json_value(parsed_field.value, action_params):
                    if len(action_params) > 0:
                        value = self.validate_json_value(record_field.type, action_params[0])
                else:
                    rf = record_types.RecordFields[record_field.type]
                    ft = record_types.FieldTypes.get(rf.type)
                    if isinstance(ft.value, str):
                        value = parsed_field.value
                        if ft.name == 'multiline':
                            value = self.validate_notes(value)
                    elif isinstance(ft.value, bool):
                        lv = parsed_field.value.lower()
                        if lv in ('1', 'y', 'yes', 't', 'true'):
                            value = True
                        elif lv in ('0', 'n', 'no', 'f', 'false'):
                            value = False
                        else:
                            self.on_warning(f'Incorrect boolean value \"{parsed_field.value}\": [t]rue or [f]alse')
                    elif isinstance(ft.value, int):
                        if parsed_field.value.isdigit():
                            value = int(parsed_field.value)
                            if value < 1_000_000_000:
                                value *= 1000
                        else:
                            if len(parsed_field.value) <= 10:
                                dt = datetime.datetime.strptime(parsed_field.value, '%Y-%m-%d')
                            else:
                                dt = datetime.datetime.strptime(parsed_field.value, '%Y-%m-%dT%H:%M:%SZ')
                            value = int(dt.timestamp() * 1000)
                    elif isinstance(ft.value, dict):
                        if ft.name == 'name':
                            value = vault.TypedField.import_name_field(parsed_field.value)
                        elif ft.name == 'address':
                            value = vault.TypedField.import_address_field(parsed_field.value)
                        elif ft.name == 'host':
                            value = vault.TypedField.import_host_field(parsed_field.value)
                        elif ft.name == 'phone':
                            value = vault.TypedField.import_phone_field(parsed_field.value)
                        elif ft.name == 'paymentCard':
                            value = vault.TypedField.import_card_field(parsed_field.value)
                        elif ft.name == 'bankAccount':
                            value = vault.TypedField.import_account_field(parsed_field.value)
                        elif ft.name == 'securityQuestion':
                            value = []
                            for qa in parsed_field.value.split(';'):
                                qa = qa.strip()
                                qav = vault.TypedField.import_q_and_a_field(qa)
                                if qav:
                                    value.append(qav)
                        elif ft.name == 'privateKey':
                            value = vault.TypedField.import_ssh_key_field(parsed_field.value)
                        elif ft.name == 'schedule':
                            value = vault.TypedField.import_schedule_field(parsed_field.value)
                        else:
                            self.on_warning(f'Unsupported field type: {record_field.type}')
                if value:
                    if isinstance(value, list):
                        record_field.value.clear()
                        record_field.value.extend(value)
                    else:
                        if len(record_field.value) == 0:
                            record_field.value.append(value)
                        else:
                            if isinstance(value, dict) and isinstance(record_field.value[0], dict):
                                record_field.value[0].update(value)
                                noneKeys = [k for k,v in record_field.value[0].items() if v is None]
                                for k in noneKeys:
                                    del record_field.value[0][k]
                            else:
                                record_field.value[0] = value
            else:
                if is_field:
                    record_field.value.clear()
                else:
                    index = next((i for i, x in enumerate(record.custom) if x is record_field), -1)
                    if 0 <= index < len(record.custom):
                        record.custom.pop(index)

    def upload_attachments(self, params, record, files, stop_on_error):
        # type: (KeeperParams, Union[vault.PasswordRecord, vault.TypedRecord], List[ParsedFieldValue], bool) -> None
        tasks = []
        for file_attachment in files:
            if file_attachment.value.startswith('@'):
                file_name = file_attachment.value[1:]
            else:
                file_name = file_attachment.value
            file_name = os.path.expanduser(file_name)
            if os.path.isfile(file_name):
                task = attachment.FileUploadTask(file_name)
                task.title = file_attachment.label
                tasks.append(task)
            else:
                self.on_warning(f'Upload attachment: file \"{file_name}\" not found')
                if stop_on_error:
                    return

        for task in tasks:
            try:
                self.on_info(f'Uploading {task.name} ...')
                attachment.upload_attachments(params, record, [task])
            except Exception as e:
                self.on_warning(str(e))
                if stop_on_error:
                    break

    def delete_attachments(self, params, record, file_names):
        # type: (KeeperParams, Union[vault.PasswordRecord, vault.TypedRecord], List[str]) -> None
        if isinstance(record, vault.PasswordRecord):
            for file_name in file_names:
                indexes = [i for i, x in enumerate(record.attachments or [])
                           if x.id == file_name or file_name.lower() in (x.name.lower(), x.title.lower())]
                if len(indexes) > 1:
                    self.on_warning(
                        f'There are multiple file attachments with name \"{file_name}\". Use attachment ID.')
                elif len(indexes) == 1:
                    record.attachments.pop(indexes[0])
        elif isinstance(record, vault.TypedRecord):
            facade = record_facades.FileRefRecordFacade()
            facade.record = record
            for file_name in file_names:
                index = next((i for i, x in enumerate(facade.file_ref) if x == file_name), -1)
                if index > 0:
                    facade.file_ref.pop(index)
                else:
                    file_uids = []
                    f_name = file_name.lower()
                    for file_uid in facade.file_ref:
                        if file_uid in params.record_cache:
                            file = vault.KeeperRecord.load(params, file_uid)
                            if isinstance(file, vault.FileRecord):
                                if f_name in (file.name.lower(), file.title.lower()):
                                    file_uids.append(file_uid)
                    if len(file_uids) > 1:
                        self.on_warning(
                            f'There are multiple file attachments with name \"{file_name}\". Use attachment ID.')
                    elif len(file_uids) == 1:
                        facade.file_ref.remove(file_uids[0])

    @staticmethod
    def get_record_type_fields(params, record_type):      # type: (KeeperParams, str) -> Optional[List[Dict]]
        rti = recordv3.RecordTypeInfo()
        js_rt = rti.execute(params, format='json', record_name=record_type)
        j_rt = json.loads(js_rt)
        if isinstance(j_rt, list):
            if len(j_rt) > 0:
                jrt_fields = j_rt[0].get('content')
                j_rt = json.loads(jrt_fields)
                if isinstance(j_rt, dict):
                    return j_rt.get('fields')


class RecordAddCommand(Command, RecordEditMixin):
    def __init__(self):
        super(RecordAddCommand, self).__init__()

    def get_parser(self):
        return record_add_parser

    def execute(self, params, **kwargs):
        if kwargs.get('syntax_help') is True:
            print(record_fields_description)
            return

        expiration_period = None
        self_destruct = kwargs.get('self_destruct')
        if self_destruct:
            expiration_period = parse_timeout(self_destruct)
            if expiration_period.total_seconds() > 182 * 24 * 60 * 60:
                raise CommandError('', 'URL expiration period cannot be greater than 6 months.')

        folder_uid = FolderMixin.resolve_folder(params, kwargs.get('folder'))

        self.warnings.clear()
        title = kwargs.get('title')
        if not title:
            raise CommandError('record-add', 'Title parameter is required.')
        record_type = kwargs.get('record_type')   # type: Optional[str]
        if not record_type:
            raise CommandError('record-add', 'Record type parameter is required.')

        fields = kwargs.get('fields', [])
        # Filter out empty strings that might be introduced by copy-paste or line continuation issues
        fields = [field.strip() for field in fields if field.strip()]

        record_fields = []    # type: List[ParsedFieldValue]
        add_attachments = []  # type: List[ParsedFieldValue]
        rm_attachments = []   # type: List[ParsedFieldValue]
        for field in fields:
            parsed_field = RecordEditMixin.parse_field(field)
            if parsed_field.type == 'file':
                (add_attachments if parsed_field.value else rm_attachments).append(parsed_field)
            else:
                record_fields.append(parsed_field)

        if record_type in ('legacy', 'general'):
            raise CommandError('record-add', 'Legacy record type is not supported anymore.')
            # record = vault.PasswordRecord()
            # self.assign_legacy_fields(record, record_fields)
        else:
            rt_fields = self.get_record_type_fields(params, record_type)
            if not rt_fields:
                raise CommandError('record-add', f'Record type \"{record_type}\" cannot be found.')
            record = vault.TypedRecord()
            record.type_name = record_type
            for rf in rt_fields:
                ref = rf.get('$ref')
                if not ref:
                    continue
                label = rf.get('label', '')
                required = rf.get('required', False)
                default_value = None
                if ref == 'appFiller':
                    if 'appFillerData' in rf:
                        default_value = rf['appFillerData']
                field = vault.TypedField.new_field(ref, default_value, label)
                if required is True:
                    field.required = True
                record.fields.append(field)
            self.assign_typed_fields(record, record_fields)

        record_uid = str(kwargs.get('record_uid', ''))
        if RecordV3.is_valid_ref_uid(record_uid):
            record.record_uid = record_uid
        record.title = title
        record.notes = self.validate_notes(kwargs.get('notes') or '')

        ignore_warnings = kwargs.get('force') is True
        if len(self.warnings) > 0:
            for warning in self.warnings:
                logging.warning(warning)
            if not ignore_warnings:
                return
        self.warnings.clear()

        if len(add_attachments) > 0:
            self.upload_attachments(params, record, add_attachments, not ignore_warnings)
            if len(self.warnings) > 0:
                for warning in self.warnings:
                    logging.warning(warning)
                if not ignore_warnings:
                    return

        record_management.add_record_to_folder(params, record, folder_uid)
        params.sync_data = True
        params.environment_variables[LAST_RECORD_UID] = record.record_uid
        if expiration_period is not None:
            record_uid = record.record_uid
            record_key = record.record_key
            client_key = utils.generate_aes_key()
            client_id = crypto.hmac_sha512(client_key, 'KEEPER_SECRETS_MANAGER_CLIENT_ID'.encode())
            rq = APIRequest_pb2.AddExternalShareRequest()
            rq.recordUid = utils.base64_url_decode(record_uid)
            rq.encryptedRecordKey = crypto.encrypt_aes_v2(record_key, client_key)
            rq.clientId = client_id
            rq.accessExpireOn = utils.current_milli_time() + int(expiration_period.total_seconds() * 1000)
            rq.isSelfDestruct = True
            api.communicate_rest(params, rq, 'vault/external_share_add', rs_type=APIRequest_pb2.Device)
            # Extract hostname from params.server in case it contains full URL with protocol
            from urllib.parse import urlparse
            parsed = urlparse(params.server)
            server_netloc = parsed.netloc if parsed.netloc else parsed.path  # parsed.path for plain hostname
            url = urlunparse(('https', server_netloc, '/vault/share', None, None, utils.base64_url_encode(client_key)))
            return url
        else:
            BreachWatch.scan_and_update_security_data(params, record.record_uid, params.breach_watch)
            return record.record_uid


class RecordUpdateCommand(Command, RecordEditMixin, RecordMixin):
    def __init__(self):
        super(RecordUpdateCommand, self).__init__()
        self.warnings = []

    def get_parser(self):
        return record_update_parser

    def execute(self, params, **kwargs):
        if kwargs.get('syntax_help') is True:
            print(record_fields_description)
            return

        self.warnings.clear()
        record_name = kwargs.get('record')
        if not record_name:
            raise CommandError('record-update', 'Record parameter is required.')
        
        from ..enforcement import MasterPasswordReentryEnforcer
        if not MasterPasswordReentryEnforcer.check_and_enforce(params, "record_level"):
            raise CommandError('record-update', 'Operation cancelled: Re-authentication failed')

        record = RecordMixin.resolve_single_record(params, record_name)
        if not record:
            raise CommandError('record-update', f'Record \"{record_name}\" not found.')
        if not isinstance(record, (vault.PasswordRecord, vault.TypedRecord)):
            raise CommandError('record-update', f'Record \"{record_name}\" can not be edited.')

        title = kwargs.get('title')
        if title:
            record.title = title
        notes = kwargs.get('notes')
        if isinstance(notes, str):
            notes = self.validate_notes(notes)
            append_notes = False
            if notes.startswith('+'):
                append_notes = True
                notes = notes[1:].strip()
            if append_notes:
                if record.notes:
                    record.notes += '\n'
                record.notes += notes
            else:
                record.notes = notes

        fields = kwargs.get('fields', [])
        # Filter out empty strings that might be introduced by copy-paste or line continuation issues
        fields = [field.strip() for field in fields if field.strip()]

        record_fields = []    # type: List[ParsedFieldValue]
        add_attachments = []  # type: List[ParsedFieldValue]
        rm_attachments = []   # type: List[ParsedFieldValue]
        for field in fields:
            parsed_field = RecordEditMixin.parse_field(field)
            if parsed_field.type == 'file':
                (add_attachments if parsed_field.value else rm_attachments).append(parsed_field)
            else:
                record_fields.append(parsed_field)

        if isinstance(record, vault.PasswordRecord):
            self.assign_legacy_fields(record, record_fields)
        elif isinstance(record, vault.TypedRecord):
            record_type = kwargs.get('record_type')
            if record_type:
                record.type_name = record_type
                rt_fields = self.get_record_type_fields(params, record_type)
                if not rt_fields:
                    raise CommandError('record-update', f'Record type \"{record_type}\" cannot be found.')
                self.adjust_typed_record_fields(record, rt_fields)
            self.assign_typed_fields(record, record_fields)
        else:
            raise CommandError('record-update', f'Record \"{record_name}\" can not be edited.')

        ignore_warnings = kwargs.get('force') is True
        if len(self.warnings) > 0:
            for warning in self.warnings:
                logging.warning(warning)
            if not ignore_warnings:
                return
        self.warnings.clear()

        if len(rm_attachments) > 0:
            names = [x.label for x in rm_attachments if x.label]
            self.delete_attachments(params, record, names)
            if len(self.warnings) > 0:
                for warning in self.warnings:
                    logging.warning(warning)
                if not ignore_warnings:
                    return
            self.warnings.clear()

        if len(add_attachments) > 0:
            self.upload_attachments(params, record, add_attachments, not ignore_warnings)
            if len(self.warnings) > 0:
                for warning in self.warnings:
                    logging.warning(warning)
                if not ignore_warnings:
                    return

        record_management.update_record(params, record)
        params.sync_data = True


class RecordAppendNotesCommand(Command):
    def get_parser(self):
        return append_parser

    def execute(self, params, **kwargs):
        notes = kwargs['notes'] if 'notes' in kwargs else None
        while not notes:
            notes = input("... Notes to append: ")

        edit_command = RecordUpdateCommand()
        kwargs['notes'] = '+' + notes
        edit_command.execute(params, **kwargs)


class RecordDeleteAttachmentCommand(Command):
    def get_parser(self):
        return delete_attachment_parser

    def execute(self, params, **kwargs):
        record_name = kwargs['record'] if 'record' in kwargs else None

        if not record_name:
            self.get_parser().print_help()
            return

        record_uid = None
        if record_name in params.record_cache:
            record_uid = record_name
        else:
            rs = try_resolve_path(params, record_name)
            if rs is not None:
                folder, record_name = rs
                if folder is not None and record_name is not None:
                    folder_uid = folder.uid or ''
                    if folder_uid in params.subfolder_record_cache:
                        for uid in params.subfolder_record_cache[folder_uid]:
                            r = api.get_record(params, uid)
                            if r.title.lower() == record_name.lower():
                                record_uid = uid
                                break

        if record_uid is None:
            raise CommandError('delete-attachment', 'Enter name or uid of existing record')

        names = kwargs['name'] if 'name' in kwargs else None
        if names is None:
            raise CommandError('delete-attachment', 'No file names')

        record = vault.KeeperRecord.load(params, record_uid)
        if not record:
            logging.warning('Record UID \"%s\" not found.', record_uid)

        deleted_files = set()
        if isinstance(record, vault.PasswordRecord):
            if record.attachments:
                for name in names:
                    for atta in record.attachments:
                        if atta.id == name:
                            deleted_files.add(atta.id)
                        elif atta.title and atta.title.lower() == name.lower():
                            deleted_files.add(atta.id)
                        elif atta.name and atta.name.lower() == name.lower():
                            deleted_files.add(atta.id)
                if len(deleted_files) > 0:
                    record.attachments = [x for x in record.attachments if x.id not in deleted_files]
        elif isinstance(record, vault.TypedRecord):
            typed_field = record.get_typed_field('fileRef')
            if typed_field and isinstance(typed_field.value, list):
                for name in names:
                    for file_uid in typed_field.value:
                        if file_uid == name:
                            deleted_files.add(file_uid)
                        else:
                            file_record = vault.KeeperRecord.load(params, file_uid)
                            if isinstance(file_record, vault.FileRecord):
                                if file_record.title.lower() == name.lower():
                                    deleted_files.add(file_uid)
                                elif file_record.name.lower() == name.lower():
                                    deleted_files.add(file_uid)
                if len(deleted_files) > 0:
                    typed_field.value = [x for x in typed_field.value if x not in deleted_files]

        if len(deleted_files) == 0:
            logging.info('Attachment not found')
            return

        record_management.update_record(params, record)
        if params.enterprise_ec_key:
            for file_uid in deleted_files:
                params.queue_audit_event('file_attachment_deleted', record_uid=record_uid, attachment_id=file_uid)
        params.sync_data = True


class RecordDownloadAttachmentCommand(Command):
    def get_parser(self):
        return download_parser

    def execute(self, params, **kwargs):
        records = kwargs.get('records')
        if not records:
            self.get_parser().print_help()
            return

        record_uids = set()
        for record in records:
            folder = None
            if record in params.record_cache:
                record_uids.add(record)
            elif record in params.folder_cache:
                folder = params.folder_cache[record]
            else:
                rs = try_resolve_path(params, record)
                if rs is not None:
                    fol, name = rs
                    if fol is not None:
                        if name:
                            f_uid = fol.uid or ''
                            if f_uid in params.subfolder_record_cache:
                                for uid in params.subfolder_record_cache[f_uid]:
                                    r = vault.KeeperRecord.load(params, uid)
                                    if isinstance(r, (vault.PasswordRecord, vault.TypedRecord)):
                                        if r.title.lower() == name.lower():
                                            record_uids.add(r.record_uid)
                        else:
                            folder = fol
            if folder:
                folders = set()
                folder_uid = folder.uid or ''
                folders.add(folder_uid)
                if kwargs.get('recursive') is True:
                    FolderMixin.traverse_folder_tree(
                        params, folder_uid, lambda x: folders.add(x.uid))
                for uid in folders:
                    if uid in params.subfolder_record_cache:
                        for record_uid in params.subfolder_record_cache[uid]:
                            if record_uid in params.record_cache:
                                record_uids.add(record_uid)

        if len(record_uids) == 0:
            logging.error('Record(s) "%s" not found', ', '.join(records))
            return
        output_dir = kwargs.get('out_dir')
        if output_dir:
            output_dir = os.path.expanduser(output_dir)
        else:
            output_dir = os.getcwd()
        if not os.path.isdir(output_dir):
            os.makedirs(output_dir)

        preserve_dir = kwargs.get('preserve_dir') is True
        record_title = kwargs.get('record_title') is True
        for record_uid in record_uids:
            attachments = list(attachment.prepare_attachment_download(params, record_uid))
            if len(attachments) == 0:
                continue

            subfolder_path = ''
            if preserve_dir:
                folder_uid = next((x for x in find_folders(params, record_uid)), None)
                if folder_uid:
                    subfolder_path = get_folder_path(params, folder_uid, os.sep)
                    subfolder_path = ''.join(x for x in subfolder_path if x.isalnum() or x == os.sep)
                    subfolder_path = subfolder_path.replace(2*os.sep, os.sep)
            if subfolder_path:
                subfolder_path = os.path.join(output_dir, subfolder_path)
                if not os.path.isdir(subfolder_path):
                    os.makedirs(subfolder_path)
            else:
                subfolder_path = output_dir

            title = ''
            if record_title:
                record = vault.KeeperRecord.load(params, record_uid)
                title = record.title
                title = ''.join(x for x in title if x.isalnum() or x.isspace())

            for atta in attachments:
                file_name = atta.title
                if title:
                    file_name = f'{title}-{atta.title}'
                file_name = os.path.basename(file_name)
                name = os.path.join(subfolder_path, file_name)
                if os.path.isfile(name):
                    base_name, ext = os.path.splitext(file_name)
                    name = os.path.join(subfolder_path, f'{base_name}({record_uid}){ext}')
                if os.path.isfile(name):
                    base_name, ext = os.path.splitext(file_name)
                    name = os.path.join(subfolder_path, f'{base_name}({atta.file_id}){ext}')
                atta.download_to_file(params, name)


class RecordUploadAttachmentCommand(Command):
    def get_parser(self):
        return upload_parser

    def execute(self, params, **kwargs):
        record_name = kwargs['record'] if 'record' in kwargs else None
        if not record_name:
            self.get_parser().print_help()
            return

        record_uid = None
        if record_name in params.record_cache:
            record_uid = record_name
        else:
            rs = try_resolve_path(params, record_name)
            if rs is not None:
                folder, record_name = rs
                if folder is not None and record_name is not None:
                    folder_uid = folder.uid or ''
                    if folder_uid in params.subfolder_record_cache:
                        for uid in params.subfolder_record_cache[folder_uid]:
                            r = api.get_record(params, uid)
                            if r.title.lower() == record_name.lower():
                                record_uid = uid
                                break

        if not record_uid:
            logging.error('Record UID not found for record "%s"', str(record_name))
            return

        upload_tasks = []
        files = kwargs.get('file')
        if isinstance(files, list):
            for name in files:
                file_name = os.path.abspath(os.path.expanduser(name))
                if os.path.isfile(file_name):
                    upload_tasks.append(attachment.FileUploadTask(file_name))
                else:
                    raise CommandError('upload-attachment', f'File "{name}" does not exists')

        if len(upload_tasks) == 0:
            raise CommandError('upload-attachment', 'No files to upload')

        record = vault.KeeperRecord.load(params, record_uid)
        if isinstance(record, (vault.PasswordRecord, vault.TypedRecord)):
            attachment.upload_attachments(params, record, upload_tasks)
            record_management.update_record(params, record)
            params.sync_data = True



#  _  __
# | |/ /___ ___ _ __  ___ _ _ ®
# | ' </ -_) -_) '_ \/ -_) '_|
# |_|\_\___\___| .__/\___|_|
#              |_|
#
# Keeper Commander
# Copyright 2025 Keeper Security Inc.
# Contact: ops@keepersecurity.com
#

from ..commands.base import GroupCommand

from .commands.register import BiometricRegisterCommand
from .commands.list import BiometricListCommand
from .commands.unregister import BiometricUnregisterCommand
from .commands.verify import BiometricVerifyCommand
from .commands.update_name import BiometricUpdateNameCommand

from .client import BiometricClient
from .platforms.detector import BiometricDetector

def check_biometric_previously_used(username):
    """Check if biometric authentication was previously used for this user"""
    try:
        detector = BiometricDetector()
        handler = detector.get_platform_handler()
        return handler.get_biometric_flag(username)
    except Exception:
        return False

class BiometricCommand(GroupCommand):
    """Main biometric command group"""
    
    def __init__(self):
        super().__init__()
        self.register_command('register', BiometricRegisterCommand(), 'Add biometric authentication method')
        self.register_command('list', BiometricListCommand(), 'List biometric authentication methods')
        self.register_command('unregister', BiometricUnregisterCommand(), 'Disable biometric authentication for this user')
        self.register_command('verify', BiometricVerifyCommand(), 'Verify biometric authentication with existing credentials')
        self.register_command('update-name', BiometricUpdateNameCommand(), 'Update friendly name of a biometric passkey')

__all__ = [
    'BiometricCommand',
    'BiometricRegisterCommand',
    'BiometricListCommand', 
    'BiometricUnregisterCommand',
    'BiometricVerifyCommand',
    'BiometricClient',
    'BiometricDetector',
    'check_biometric_previously_used',
]



#  _  __
# | |/ /___ ___ _ __  ___ _ _ ®
# | ' </ -_) -_) '_ \/ -_) '_|
# |_|\_\___\___| .__/\___|_|
#              |_|
#
# Keeper Commander
# Copyright 2021 Keeper Security Inc.
# Contact: ops@keepersecurity.com
#

import base64
import json
import logging
import math
import re
import time
from urllib.parse import urlparse, parse_qs, unquote
from pathlib import Path
import sys

from . import crypto
from .constants import EMAIL_PATTERN


VALID_URL_SCHEME_CHARS = '+-.:'

def get_logger():
    return logging.getLogger('keeper')

def get_default_path():
    default_path = Path.home().joinpath('.keeper')
    default_path.mkdir(parents=True, exist_ok=True)
    return default_path


def generate_uid():             # type: () -> str
    b = crypto.get_random_bytes(16)
    if (b[0] & 0xf8) == 0xf8:
        b = bytes([b[0] & 0x7f]) + b[1:]
    return base64_url_encode(b)


def generate_aes_key():         # type: () -> bytes
    return crypto.get_random_bytes(32)


def current_milli_time():       # type: () -> int
    return int(round(time.time() * 1000))


def base64_url_decode(s):       # type: (str) -> bytes
    return base64.urlsafe_b64decode(s + '==')


def base64_url_encode(b):       # type: (bytes) -> str
    bs = base64.urlsafe_b64encode(b)
    return bs.rstrip(b'=').decode('ascii')


def string_to_bytes(string):
    return string.encode('utf-8')


def json_to_base64(json_str):
    json_bytes = string_to_bytes(json_str)
    json_b64 = base64.b64encode(json_bytes).decode()
    return json_b64


def decrypt_encryption_params(encryption_params, password):     # type: (bytes, str) -> bytes

    if len(encryption_params) != 100:
        raise Exception('Invalid encryption params: bad params length')

    _ = int.from_bytes(encryption_params[0:1], byteorder='big', signed=False)
    iterations = int.from_bytes(encryption_params[1:4], byteorder='big', signed=False)
    salt = encryption_params[4:20]
    encrypted_data_key = encryption_params[20:]

    key = crypto.derive_key_v1(password, salt, iterations)
    decrypted_data_key = crypto.decrypt_aes_v1(encrypted_data_key, key, use_padding=False)

    # validate the key is formatted correctly
    if len(decrypted_data_key) != 64:
        raise Exception('Invalid data key length')

    if decrypted_data_key[:32] != decrypted_data_key[32:]:
        raise Exception('Invalid data key: failed mirror verification')

    return decrypted_data_key[:32]


def create_encryption_params(password, salt, iterations, data_key):  # type: (str, bytes, int, bytes) -> bytes

    key = crypto.derive_key_v1(password, salt, iterations)
    enc_iter = int.to_bytes(iterations, length=3, byteorder='big', signed=False)
    enc_iv = crypto.get_random_bytes(16)
    enc_data_key = crypto.encrypt_aes_v1(data_key * 2, key, iv=enc_iv, use_padding=False)
    return b'\x01' + enc_iter + salt + enc_data_key


def create_auth_verifier(password, salt, iterations):   # type: (str, bytes, int) -> bytes

    derived_key = crypto.derive_key_v1(password, salt, iterations)
    enc_iter = int.to_bytes(iterations, length=3, byteorder='big', signed=False)
    return b'\x01' + enc_iter + salt + derived_key


def is_url(test_str):   # type: (str) -> bool
    if not isinstance(test_str, str):
        return False
    url_parts = test_str.split('://')
    url_scheme = url_parts[0]
    valid_scheme = all(c.isalnum or c in VALID_URL_SCHEME_CHARS for c in url_scheme)
    if len(test_str.split()) == 1 and len(url_parts) > 1 and valid_scheme:
        return True
    else:
        return False


email_pattern = re.compile(EMAIL_PATTERN)


def is_email(test_str):
    return email_pattern.match(test_str) is not None


def is_json(txt):
    try:
        json.loads(txt)
    except ValueError as e:
        return False
    return True

def url_strip(url):   # type: (str) -> str
    if not url:
        return ''
    try:
        result = urlparse(url)
        return result.netloc + result.path
    except Exception:
        return ''


_breach_watch_key = base64_url_decode('phl9kdMA_gkJkSfeOYWpX-FOyvfh-APhdSFecIDMyfI')


def breach_watch_hash(password):  # type: (str) -> bytes
    return crypto.hmac_sha512(_breach_watch_key, f'password:{password}'.encode('utf-8'))


def chunk_text(text, func):  # type (str, Callable[[str], bool]) -> Iterable[str]
    acc = ''
    for x in text:
        if func(x):
            acc += x
        else:
            if acc:
                yield acc
                acc = ''
    if acc:
        yield acc


def offset_char(text, func):  # type (str, Callable[[str, str], int]) -> Iterable[int]
    if not text:
        return
    prev = text[0]
    for ch in text[1:]:
        yield func(prev, ch)
        prev = ch


def password_score(password):  # type: (str) -> int
    score = 0
    if not password:
        return score
    if not isinstance(password, str):
        return score

    total = len(password)
    uppers = 0
    lowers = 0
    digits = 0
    symbols = 0
    for x in password:
        if x.isupper():
            uppers += 1
        elif x.islower():
            lowers += 1
        elif x.isdecimal():
            digits += 1
        else:
            symbols += 1

    ds = digits + symbols
    if not password[0].isalpha():
        ds -= 1
    if not password[-1].isalpha():
        ds -= 1
    if ds < 0:
        ds = 0

    score += total * 4
    if uppers > 0:
        score += (total-uppers) * 2
    if lowers > 0:
        score += (total-lowers) * 2
    if digits > 0:
        score += digits * 4
    score += symbols * 6
    score += ds * 2

    variance = 0
    if uppers > 0:
        variance += 1
    if lowers > 0:
        variance += 1
    if digits > 0:
        variance += 1
    if symbols > 0:
        variance += 1
    if total >= 8 and variance >= 3:
        score += (variance + 1) * 2

    if digits + symbols == 0:
        score -= total

    if uppers + lowers + symbols == 0:
        score -= total

    rep_inc = 0
    pwd_len = len(password)
    rep_count = 0
    for i in range(pwd_len):
        char_exists = False
        for j in range(pwd_len):
            if i != j and password[i] == password[j]:
                char_exists = True
                rep_inc += pwd_len / abs(i - j)
        if char_exists:
            rep_count += 1
            unq_count = pwd_len - rep_count
            rep_inc = math.ceil(rep_inc if unq_count == 0 else rep_inc / unq_count)

    if rep_count > 0:
        score -= rep_inc

    count = 0
    for consec in [str.isupper, str.islower, str.isdecimal]:
        for chunk in chunk_text(password, consec):
            length = len(chunk)
            if length >= 2:
                count += length - 1
    if count > 0:
        score -= 2 * count

    count = 0
    for cnt, seq in [(26, str.isalpha), (10, str.isdecimal)]:
        cnt = 0
        for chunk in chunk_text(password.lower(), seq):
            if len(chunk) >= 3:
                offsets = [x if x >= 0 else x + cnt for x in offset_char(chunk, lambda x, y: ord(x) - ord(y))]
                op = offsets[0]
                for oc in offsets[1:]:
                    if oc == op:
                        if op != 0:
                            count += 1
                    else:
                        op = oc

    symbols = {x[1]: x[0] for x in enumerate('!@#$%^&*()_+[]\\{}|;\':\",./<>?')}
    cnt = 0
    for chunk in chunk_text(password, symbols.__contains__):
        if len(chunk) >= 3:
            offsets = [x if x >= 0 else x + cnt for x in offset_char(chunk, lambda x, y: symbols[x] - symbols[y])]
            op = offsets[0]
            for oc in offsets[1:]:
                if oc == op:
                    if op != 0:
                        count += 1
                else:
                    op = oc

    if count > 0:
        score -= 3 * count

    return score if 0 <= score <= 100 else 0 if score < 0 else 100


def is_pw_weak(pw_score):           # type: (int) -> bool
    return pw_score < 40


def is_pw_fair(pw_score):           # type: (int) -> bool
    return 40 <= pw_score < 60


def is_pw_strong(pw_score):         # type: (int) -> bool
    return pw_score >= 80


def is_rec_at_risk(bw_result):      # type (int) -> bool
    return bw_result in (2, 3)


def passed_bw_check(bw_result):     # type (int) -> bool
    return bw_result in (0, 1)


def confirm(msg):
    """Simple confirmation through user input

    msg(str): Message expecting a yes or no answer
    Returns True if answer is "yes" and False otherwise.
    """
    question = f'{msg} (y/n) '
    answer = ''
    while answer not in ('y', 'n'):
        answer = input(question).lower()
    return answer == 'y'


def size_to_str(size):  # type: (int) -> str
    if not isinstance(size, int):
        return ''
    if size < 2000:
        return f'{size} b'
    size = size / 1024
    if size < 1000:
        return f'{size:.2f} Kb'
    size = size / 1024
    if size < 1000:
        return f'{size:.2f} Mb'
    size = size / 1024
    return f'{size:,.2f} Gb'

def parse_totp_uri(uri):    # type: (str) -> Dict[str, Union[str, int, None]]
    def parse_int(val):
        return val and int(val)

    def decode_uri_component(component):  # type: (str) -> str
        return unquote(component or '').strip()

    result = dict()

    if not uri:
        return result

    parsed = urlparse(uri)
    if parsed.scheme == 'otpauth':
        label = re.sub(r'^/+', '', parsed.path or '')
        parts = re.split(r':|%3A', label)
        parts = [part for part in parts if part]
        account_name = len(parts) and parts.pop()
        issuer = len(parts) and parts.pop()

        parsed = parse_qs(parsed.query)

        issuers = parsed.get('issuer')
        secrets = parsed.get('secret')
        algorithms = parsed.get('algorithm')
        digits_vals = parsed.get('digits')
        periods = parsed.get('period')
        result = {
            'issuer': decode_uri_component(issuers and next(iter(issuers)) or issuer),
            'account': decode_uri_component(account_name),
            'secret': secrets and next(iter(secrets)),
            'algorithm': algorithms and next(iter(algorithms)) or 'SHA1',
            'digits': parse_int(digits_vals and next(iter(digits_vals))) or 6,
            'period': parse_int(periods and next(iter(periods))) or 30
        }

    return result

def value_to_boolean(value):
    """
    Replacement for distutils.util.strtobool
    """
    value = str(value)
    if value.lower() in ['true', 'yes', 'on', '1']:
        return True
    elif value.lower() in ['false', 'no', 'off', '0']:
        return False
    else:
        return None

def get_ssl_cert_file():
    """Get SSL certificate file path, preferring system CA store for corporate environments like Zscaler"""
    import ssl
    import platform
    import certifi
    import os
    
    # Allow user to override via environment variable
    user_cert_file = os.getenv('KEEPER_SSL_CERT_FILE')
    if user_cert_file:
        if user_cert_file.lower() == 'system':
            pass  # Continue with system detection below
        elif user_cert_file.lower() == 'certifi':
            return certifi.where()
        elif user_cert_file.lower() == 'none' or user_cert_file.lower() == 'false':
            return False  # Disable SSL verification
        elif os.path.exists(user_cert_file):
            return user_cert_file
        else:
            # Don't use logging here as it can interfere with main logging config
            print(f"Warning: SSL cert file specified in KEEPER_SSL_CERT_FILE not found: {user_cert_file}", file=sys.stderr)
    
    # Try to use system CA store first for corporate environments
    try:
        # On macOS, try Homebrew certificates first (better for corporate environments like Zscaler)
        if platform.system() == 'Darwin':
            system_ca_paths = [
                '/opt/homebrew/etc/ca-certificates/cert.pem',  # Homebrew CA bundle (best for Zscaler)
                '/usr/local/etc/ssl/cert.pem',  # Homebrew SSL (older location)
                '/etc/ssl/cert.pem',  # macOS system CA bundle
            ]
            for ca_path in system_ca_paths:
                if os.path.exists(ca_path):
                    return ca_path
        
        # On Linux/Unix systems
        elif platform.system() == 'Linux':
            system_ca_paths = [
                '/etc/ssl/certs/ca-certificates.crt',  # Debian/Ubuntu
                '/etc/pki/tls/certs/ca-bundle.crt',    # RHEL/CentOS
                '/etc/ssl/ca-bundle.pem',              # OpenSUSE
                '/etc/ssl/cert.pem',                   # Generic
            ]
            for ca_path in system_ca_paths:
                if os.path.exists(ca_path):
                    return ca_path
        
        # Try to get default SSL context locations
        try:
            default_locations = ssl.get_default_verify_paths()
            if default_locations.cafile and os.path.exists(default_locations.cafile):
                return default_locations.cafile
            if default_locations.capath and os.path.exists(default_locations.capath):
                return default_locations.capath
        except:
            pass
            
    except Exception:
        pass
    
    # Fall back to certifi if system CA not available
    return certifi.where()


def ssl_aware_request(method, url, **kwargs):
    """Make an SSL-aware HTTP request using system CA certificates when available"""
    import requests
    
    # Only set verify if not already specified
    if 'verify' not in kwargs:
        cert_file = get_ssl_cert_file()
        if cert_file is False:
            kwargs['verify'] = False
        elif cert_file:
            kwargs['verify'] = cert_file
        # If cert_file is None, let requests use its default
    
    return requests.request(method, url, **kwargs)


def ssl_aware_get(url, **kwargs):
    """SSL-aware GET request using system CA certificates when available"""
    return ssl_aware_request('GET', url, **kwargs)

def is_windows_11():
    if sys.platform != "win32":
        return False
    version = sys.getwindowsversion()
    return version.major >= 10 and version.build >= 22000


#  _  __
# | |/ /___ ___ _ __  ___ _ _ ®
# | ' </ -_) -_) '_ \/ -_) '_|
# |_|\_\___\___| .__/\___|_|
#              |_|
#
# Keeper Commander
# Copyright 2021 Keeper Security Inc.
# Contact: ops@keepersecurity.com
#

import argparse
import json
import logging
import re
from urllib.parse import urlunparse

from .base import RecordMixin, FolderMixin, raise_parse_exception, suppress_exit
from .enterprise import EnterpriseCommand
from .register import OneTimeShareCreateCommand
from .. import api, crypto, utils, generator, rest_api, vault, record_management, vault_extensions, constants
from ..constants import EMAIL_PATTERN
from ..error import CommandError
from ..loginv3 import LoginV3API
from ..proto import enterprise_pb2


def register_commands(commands):
    commands['create-user'] = CreateEnterpriseUserCommand()
    commands['store-user-keys'] = StoreUserKeysCommand()


def register_command_info(_, command_info):
    command_info['create-user'] = 'Create an Enterprise User with a reserved domain'


register_parser = argparse.ArgumentParser(prog='create-user', description='Creates enterprise user')
register_parser.add_argument('--name', dest='name', action='store', help='user name')
register_parser.add_argument('--node', dest='node', action='store', help='node name or node ID')
register_parser.add_argument('--folder', dest='folder', action='store', help='folder name or UID to store password record')
register_parser.add_argument('-v', '--verbose', dest='verbose', action='store_true', help='print verbose information')
register_parser.add_argument('email', help='email')
register_parser.error = raise_parse_exception
register_parser.exit = suppress_exit


store_user_keys_parser = argparse.ArgumentParser(prog='store-user-keys', description='Stores user keys')


class CreateEnterpriseUserCommand(EnterpriseCommand, RecordMixin):
    def get_parser(self):
        return register_parser

    def execute(self, params, **kwargs):
        node_name = kwargs.get('node')
        nodes = list(self.resolve_nodes(params, node_name))
        if len(nodes) == 0:
            logging.warning('Node \"%s\" is not found', node_name)
            return
        if len(nodes) > 1:
            logging.warning('More than one nodes \"%s\" are found', node_name)
            return
        node_id = nodes[0]['node_id']

        email = kwargs.get('email')
        if not email:
            logging.warning('Email parameter is required..')
        email_pattern = re.compile(EMAIL_PATTERN)
        match = email_pattern.match(email)
        if not match:
            logging.warning('"%s" appears not a valid email address. Skipping.', email)
            return

        tree_key = params.enterprise['unencrypted_tree_key']
        displayname = kwargs.get('name') or ''

        rq = enterprise_pb2.EnterpriseUsersProvisionRequest()
        rq.clientVersion = rest_api.CLIENT_VERSION
        data = {'displayname': displayname or email}
        user_data = json.dumps(data).encode('utf-8')
        user_password = generator.generate(20)
        user_data_key = utils.generate_aes_key()

        enterprise_user_id = EnterpriseCommand.get_enterprise_id(params)

        user_rq = enterprise_pb2.EnterpriseUsersProvision()
        user_rq.enterpriseUserId = enterprise_user_id
        user_rq.username = email
        user_rq.nodeId = node_id
        user_rq.encryptedData = utils.base64_url_encode(crypto.encrypt_aes_v1(user_data, tree_key))
        user_rq.keyType = enterprise_pb2.ENCRYPTED_BY_DATA_KEY
        enterprise_ec_key = params.enterprise_ec_key
        if 'keys' in params.enterprise:
            keys = params.enterprise['keys']
            if 'ecc_public_key' in keys:
                pub_key = utils.base64_url_decode(keys['ecc_public_key'])
                enterprise_ec_key = crypto.load_ec_public_key(pub_key)
        user_rq.enterpriseUsersDataKey = crypto.encrypt_ec(user_data_key, enterprise_ec_key)
        user_rq.authVerifier = utils.create_auth_verifier(
            user_password, crypto.get_random_bytes(16), constants.PBKDF2_ITERATIONS)
        user_rq.encryptionParams = utils.create_encryption_params(
            user_password, crypto.get_random_bytes(16), constants.PBKDF2_ITERATIONS, user_data_key)
        if not params.forbid_rsa:
            rsa_private_key, rsa_public_key = crypto.generate_rsa_key()
            rsa_private = crypto.unload_rsa_private_key(rsa_private_key)
            rsa_public = crypto.unload_rsa_public_key(rsa_public_key)
            user_rq.rsaPublicKey = rsa_public
            user_rq.rsaEncryptedPrivateKey = crypto.encrypt_aes_v1(rsa_private, user_data_key)

        ec_private_key, ec_public_key = crypto.generate_ec_key()
        ec_private = crypto.unload_ec_private_key(ec_private_key)
        ec_public = crypto.unload_ec_public_key(ec_public_key)
        user_rq.eccPublicKey = ec_public
        user_rq.eccEncryptedPrivateKey = crypto.encrypt_aes_v2(ec_private, user_data_key)

        user_rq.encryptedDeviceToken = LoginV3API.get_device_id(params)
        user_rq.encryptedClientKey = crypto.encrypt_aes_v1(utils.generate_aes_key(), user_data_key)

        rq.users.append(user_rq)

        rs = api.communicate_rest(params, rq, 'enterprise/enterprise_user_provision',
                                  rs_type=enterprise_pb2.EnterpriseUsersProvisionResponse)
        for user_rs in rs.results:
            if user_rs.code and user_rs.code not in ['success', 'ok']:
                email_provisioning_doc = 'https://docs.keeper.io/enterprise-guide/user-and-team-provisioning/email-auto-provisioning'
                raise CommandError(
                    '', f'Failed to auto-create account "{email}".\nCreating user accounts without email '
                        'verification is only permitted on reserved domains.\n'
                        'To reserve a domain please contact Keeper support. '
                        f'Learn more about domain reservation here:\n{email_provisioning_doc}')
        api.query_enterprise(params)

        login_facade = vault_extensions.LoginFacade()
        ots_command = OneTimeShareCreateCommand()

        folder_uid = None
        folder_name = kwargs.get('folder')
        if folder_name:
            folder_uid = FolderMixin.resolve_folder(params, folder_name)

        # Extract hostname from params.server in case it contains full URL with protocol
        from urllib.parse import urlparse
        parsed = urlparse(params.server)
        server_netloc = parsed.netloc if parsed.netloc else parsed.path  # parsed.path for plain hostname
        keeper_url = urlunparse(('https', server_netloc, '/vault', None, None, f'email/{email}'))
        record = vault.TypedRecord()
        login_facade.assign_record(record)
        login_facade.title = f'Keeper Account: {email}'
        login_facade.login = email
        login_facade.password = user_password
        login_facade.url = keeper_url
        login_facade.notes = 'The user is required to change their Master Password upon login.'
        record_management.add_record_to_folder(params, record, folder_uid=folder_uid)
        api.sync_down(params)
        ots_url = ots_command.execute(params, record=record.record_uid, share_name=f'{email}: Master Password', expire='7d')
        if ots_url:
            record.custom.append(vault.TypedField.new_field('url', ots_url, 'One-Time Share'))
            record_management.update_record(params, record)
            api.sync_down(params)

        if kwargs.get('verbose') is True:
            print(f'The account {login_facade.login} has been created. Login details below:')
            print(f'{"Vault Login URL:":>24s} {login_facade.url}')
            print(f'{"Email:":>24s} {login_facade.login}')
            if displayname:
                print(f'{"Name:":>24s} {displayname}')
            if len(nodes) > 0:
                node_name = (nodes[0].get('data') or {}).get('displayname') or ''
                if node_name:
                    print(f'{"Node:":>24s} {node_name}')
            print(f'{"Master Password:":>24s} {login_facade.password}')
            if ots_url:
                print(f'{"One-Time Share Link:":>24s} {ots_url}')
            print(f'{"Note:":>24s} {login_facade.notes}')
        else:
            logging.info('User \"%s\" credentials are stored to record \"%s\"', login_facade.login, login_facade.title)
            return record.record_uid


class StoreUserKeysCommand(EnterpriseCommand):
    def get_parser(self):
        return store_user_keys_parser

    def execute(self, params, **kwargs):

        user_lookup = {x['enterprise_user_id']: x['username'] for x in params.enterprise['users']}
        tree_key = params.enterprise['unencrypted_tree_key']
        rs = api.communicate_rest(params, None, 'enterprise/get_transfer_keys',
                                  rs_type=enterprise_pb2.EnterpriseUserDataKeys)
        rqs = []
        for key in rs.keys:
            if not key.roleKey:
                continue
            if not key.userEncryptedDataKey:
                continue

            user_id = key.enterpriseUserId
            username = user_lookup.get(user_id) or str(user_id)
            try:
                decrypted_role_key = crypto.decrypt_aes_v2(key.roleKey, tree_key)
                decrypted_private_key = crypto.decrypt_aes_v1(key.privateKey, decrypted_role_key)
                private_key = crypto.load_rsa_private_key(decrypted_private_key)
                user_data_key = crypto.decrypt_rsa(key.userEncryptedDataKey, private_key)

                user_key = enterprise_pb2.EnterpriseUserDataKey()
                user_key.enterpriseUserId = user_id
                user_key.userEncryptedDataKey = crypto.encrypt_ec(user_data_key, params.enterprise_ec_key)
                user_key.keyTypeId = 4
                rqs.append(user_key)
                logging.info('Converting key for user "%s"', username)
            except Exception as e:
                logging.warning('Error converting key for user %s: %s', username, e)

        while len(rqs) > 0:
            chunk = rqs[:900]
            rqs = rqs[900:]
            set_rq = enterprise_pb2.EnterpriseUserDataKeys()
            set_rq.keys.extend(chunk)
            api.communicate_rest(params, set_rq, 'enterprise/set_enterprise_users_data_keys')



from __future__ import annotations
import argparse
import logging
from . import PAMGatewayActionDiscoverCommandBase, GatewayContext
from ..pam.pam_dto import GatewayActionDiscoverRuleValidateInputs, GatewayActionDiscoverRuleValidate, GatewayAction
from ..pam.router_helper import router_send_action_to_gateway, router_get_connected_gateways
from ...display import bcolors
from ...proto import pam_pb2
from ...discovery_common.rule import Rules
from ...discovery_common.types import ActionRuleItem
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from ...params import KeeperParams


class PAMGatewayActionDiscoverRuleAddCommand(PAMGatewayActionDiscoverCommandBase):
    parser = argparse.ArgumentParser(prog='pam-action-discover-rule-add')
    parser.add_argument('--gateway', '-g', required=True, dest='gateway', action='store',
                        help='Gateway name of UID.')
    parser.add_argument('--action', '-a', required=True, choices=['add', 'ignore', 'prompt'],
                        dest='rule_action', action='store', help='Action to take if rule matches')
    parser.add_argument('--priority', '-p', required=True, dest='priority', action='store', type=int,
                        help='Rule execute priority')
    parser.add_argument('--ignore-case', required=False, dest='ignore_case', action='store_true',
                        help='Ignore value case. Rule value must be in lowercase.')
    parser.add_argument('--shared-folder-uid', required=False, dest='shared_folder_uid',
                        action='store', help='Folder to place record.')
    parser.add_argument('--statement', '-s', required=True, dest='statement', action='store',
                        help='Rule statement')

    def get_parser(self):
        return PAMGatewayActionDiscoverRuleAddCommand.parser

    @staticmethod
    def validate_rule_statement(params: KeeperParams, gateway_context: GatewayContext, statement: str):

        # Send rule the gateway to be validated. The rule is encrypted. It might contain sensitive information.
        action_inputs = GatewayActionDiscoverRuleValidateInputs(
            configuration_uid=gateway_context.configuration_uid,
            statement=gateway_context.encrypt_str(statement)
        )
        conversation_id = GatewayAction.generate_conversation_id()
        router_response = router_send_action_to_gateway(
            params=params,
            gateway_action=GatewayActionDiscoverRuleValidate(
                inputs=action_inputs,
                conversation_id=conversation_id),
            message_type=pam_pb2.CMT_DISCOVERY,
            is_streaming=False,
            destination_gateway_uid_str=gateway_context.gateway_uid
        )

        data = PAMGatewayActionDiscoverCommandBase.get_response_data(router_response)

        if data is None:
            raise Exception("The router returned a failure.")
        elif data.get("success") is False:
            error = data.get("error")
            raise Exception(f"The rule does not appear valid: {error}")

        statement_struct = data.get("statementStruct")
        logging.debug(f"Rule Structure = {statement_struct}")
        if isinstance(statement_struct, list) is False:
            raise Exception(f"The structured rule statement is not a list.")

        return statement_struct

    def execute(self, params, **kwargs):

        if not hasattr(params, 'pam_controllers'):
            router_get_connected_gateways(params)

        try:
            gateway = kwargs.get("gateway")
            gateway_context = GatewayContext.from_gateway(params, gateway)
            if gateway_context is None:
                print(f'{bcolors.FAIL}Discovery job gateway [{gateway}] was not found.{bcolors.ENDC}')
                return

            # If we are setting the shared_folder_uid, make sure it exists.
            shared_folder_uid = kwargs.get("shared_folder_uid")
            if shared_folder_uid is not None:
                shared_folder_uids = gateway_context.get_shared_folders(params)
                exists = next((x for x in shared_folder_uids if x["uid"] == shared_folder_uid), None)
                if exists is None:
                    print(f"{bcolors.FAIL}The shared folder UID {shared_folder_uid} is not part of this "
                          f"application/gateway. Valid shared folder UID are:{bcolors.ENDC}")
                    for item in shared_folder_uids:
                        print(f"* {item['uid']} - {item['name']}")
                    return

            statement = kwargs.get("statement")
            statement_struct = self.validate_rule_statement(
                params=params,
                gateway_context=gateway_context,
                statement=statement
            )

            # If the rule passes its validation, then add control DAG
            rules = Rules(record=gateway_context.configuration, params=params)
            new_rule = ActionRuleItem(
                action=kwargs.get("rule_action"),
                priority=kwargs.get("priority"),
                case_sensitive=not kwargs.get("ignore_case", False),
                shared_folder_uid=kwargs.get("shared_folder_uid"),
                statement=statement_struct,
                enabled=True
            )
            rules.add_rule(new_rule)

            print(f"{bcolors.OKGREEN}Rule has been added{bcolors.ENDC}")
        except Exception as err:
            print(f"{bcolors.FAIL}Rule was not added: {err}{bcolors.ENDC}")



#  _  __
# | |/ /___ ___ _ __  ___ _ _ ®
# | ' </ -_) -_) '_ \/ -_) '_|
# |_|\_\___\___| .__/\___|_|
#              |_|
#
# Keeper Commander
# Copyright 2022 Keeper Security Inc.
# Contact: ops@keepersecurity.com
#

import abc
from typing import TypeVar, Generic, Iterable, Optional, Tuple, Union


K = TypeVar('K', int, str, bytes)

class IUid(Generic[K], abc.ABC):
    @abc.abstractmethod
    def uid(self) -> K:
        pass

KS = TypeVar('KS', int, str, bytes)
KO = TypeVar('KO', int, str, bytes)

class IUidLink(Generic[KS, KO], abc.ABC):
    @abc.abstractmethod
    def subject_uid(self) -> KS:
        pass

    @abc.abstractmethod
    def object_uid(self) -> KO:
        pass


class UidLink(IUidLink[str, str]):
    def __init__(self, subject_uid: str, object_uid: str):
        self._subject_uid = subject_uid
        self._object_uid = object_uid

    def subject_uid(self) -> str:
        return self._subject_uid

    def object_uid(self) -> str:
        return self._object_uid

T = TypeVar('T')

class IRecordStorage(Generic[T], abc.ABC):
    @abc.abstractmethod
    def load(self) -> None:
        pass

    @abc.abstractmethod
    def store(self, record: T):
        pass

    @abc.abstractmethod
    def delete(self) -> None:
        pass

class IEntityReader(Generic[T, K], abc.ABC):
    @abc.abstractmethod
    def get_all_entities(self) -> Iterable[T]:
        pass

    @abc.abstractmethod
    def get_entity(self, key: K) -> Optional[T]:
        pass

    def get_all(self) -> Iterable[T]:
        return self.get_all_entities()


class IEntityStorage(IEntityReader[T, K], abc.ABC):
    @abc.abstractmethod
    def put_entities(self, entities: Iterable[T]) -> None:
        pass

    @abc.abstractmethod
    def delete_uids(self, uids: Iterable[K]) -> None:
        pass

class ILinkReader(Generic[T, KS, KO], abc.ABC):
    @abc.abstractmethod
    def get_link(self, subject_id: KS, object_id: KO) -> Optional[T]:
        pass

    @abc.abstractmethod
    def get_links_for_subject(self, subject_id: KS) -> Iterable[T]:
        pass

    @abc.abstractmethod
    def get_links_for_object(self, object_id: KO) -> Iterable[T]:
        pass

    @abc.abstractmethod
    def get_all_links(self) -> Iterable[T]:
        pass

class ILinkStorage(ILinkReader[T, KS, KO], abc.ABC):
    @abc.abstractmethod
    def put_links(self, links: Iterable[T]) -> None:
        pass

    @abc.abstractmethod
    def delete_links(self, links: Iterable[Union[Tuple[KS, KO], IUidLink[KS, KO]]]) -> None:
        pass

    @abc.abstractmethod
    def delete_links_for_subjects(self, subject_uids: Iterable[KS]) -> None:
        pass

    @abc.abstractmethod
    def delete_links_for_objects(self, object_uids: Iterable[KO]) -> None:
        pass



from __future__ import annotations
from pydantic import BaseModel
from typing import Optional, List


class RmResponse(BaseModel):
    """
    Generic response for action.

    :param notes: A list of notes on what happened in the action.
    """
    notes: List[str] = []


class RmScriptResponse(BaseModel):
    """
    The response from running a script

    :param script: The script with all placeholders replaced.
    :param stdout: The STDOUT from running the script.
    :param stderr: The STDERR from running the script.
    """
    script: str
    stdout: Optional[str] = None
    stderr: Optional[str] = None


class RmKeyValue(BaseModel):
    """
    A key/values pair.

    :param key: The key for the value.
    :param value: The value stored a string.
    """
    key: str
    value: str


class RmInformation(BaseModel):
    name: str
    record_type: str
    version: Optional[str] = None
    version_number: Optional[str] = None
    supports_groups: bool = False
    supports_roles: bool = False
    can_create_users: bool = False
    can_delete_users: bool = False
    can_run_scripts: bool = False
    settings: List[RmKeyValue] = []


class RmUser(BaseModel):
    id: str
    name: str
    desc: Optional[str] = None


class RmNewUser(BaseModel):
    id: str
    name: str
    desc: Optional[str] = None
    password: Optional[str] = None
    private_key: Optional[str] = None
    private_key_passphrase: Optional[str] = None
    connect_database: Optional[str] = None
    dn: Optional[str] = None


class RmRole(BaseModel):
    id: str
    name: Optional[str] = None
    desc: Optional[str] = None
    users: List[RmUser] = []


class RmGroup(BaseModel):
    id: str
    name: Optional[str] = None
    desc: Optional[str] = None
    users: List[RmUser] = []


class RmMetaBase(BaseModel):
    pass


class RmStubMeta(RmMetaBase):
    pass


class RmAwsUserAddMeta(RmMetaBase):
    console_access:  Optional[bool] = True
    path: Optional[str] = "/"
    permission_boundary_arn: Optional[str] = None
    password_reset_required: Optional[bool] = False
    tags: List[RmKeyValue] = []
    roles: List[str] = []
    groups: List[str] = []
    policies: List[str] = []


class RmAwsRoleAddMeta(RmMetaBase):
    path: Optional[str] = None
    trust_policy_json: Optional[str] = None
    duration_seconds: Optional[int] = None
    attach_policy_arns: List[str] = []
    permission_boundary_arn: Optional[str] = None
    description: Optional[str] = None
    tags: List[RmKeyValue] = []


class RmAwsGroupAddMeta(RmMetaBase):
    path: Optional[str] = None
    inline_policy_json: List[str] = []
    attach_policy_arns: List[str] = []


class RmAzureUserAddMeta(RmMetaBase):
    account_enabled: Optional[bool] = True
    display_name: Optional[str] = None
    on_premise_immutable_id: Optional[str] = None
    password_reset_required: Optional[bool] = False
    password_reset_required_with_mfa: Optional[bool] = False
    roles: List[str] = []
    groups: List[str] = []


class RmAzureRoleAddMeta(RmMetaBase):
    custom_role_json: str
    description: Optional[str] = None


class RmAzureGroupAddMeta(RmMetaBase):
    mail_enabled: bool = False
    mail_nickname: Optional[str] = None
    security_enabled: bool = True
    group_types: List[str] = []


class RmDomainUserAddMeta(RmMetaBase):
    roles: List[str] = []
    groups: List[str] = []

# DATABASE


class RmMySQLUserAddMeta(RmMetaBase):
    authentication_plugin: Optional[str] = None
    authentication_value: Optional[str] = None
    roles: List[str] = []


class RmMySQLRoleAddMeta(RmMetaBase):
    grant_script: Optional[str] = None


class RmMariaDbRoleAddMeta(RmMetaBase):
    with_admin: str = "CURRENT_USER"
    grant_script: Optional[str] = None


class RmMariaDbLUserAddMeta(RmMetaBase):
    authentication_plugin: Optional[str] = None
    authentication_value: Optional[str] = None
    roles: List[str] = []


class RmPostgreSqlUserAddMeta(RmMetaBase):
    superuser: Optional[bool] = False
    create_db: Optional[bool] = False
    create_role: Optional[bool] = False
    inherit: Optional[bool] = False
    login: Optional[bool] = True
    replication: Optional[bool] = False
    bypass_rls: Optional[bool] = False
    connection_limit: Optional[int] = None
    valid_until: Optional[str] = None
    roles: List[str] = []
    inc_in_roles: List[str] = []
    inc_in_roles_as_admin: List[str] = []
    sysid: Optional[str] = None


class RmPostgreSqlRoleAddMeta(RmMetaBase):

    # Same as users
    superuser: Optional[bool] = False
    create_db: Optional[bool] = False
    create_role: Optional[bool] = False
    inherit: Optional[bool] = False
    login: Optional[bool] = False
    replication: Optional[bool] = False
    bypass_rls: Optional[bool] = False
    connection_limit: Optional[int] = None
    valid_until: Optional[str] = None
    roles: List[str] = []
    inc_in_roles: List[str] = []
    inc_in_roles_as_admin: List[str] = []
    sysid: Optional[str] = None

    # SQL to creat GRANTS
    grant_script: Optional[str] = None


class RmSqlServerUserAddMeta(RmMetaBase):
    allow_login: bool = True
    use_windows_auth: bool = False
    is_reader: bool = True
    is_writer: bool = True
    roles: List[str] = []


class RmSqlServerRoleAddMeta(RmMetaBase):
    """
    Meta information for adding a Sql Server role.

    :param authorization: The login that will own the new server role.
                          If no login is specified, the server role will be owned by the login that executes
                            CREATE SERVER ROLE.
    :param grant_script: The STDOUT from running the script.
    """

    authorization: Optional[str] = None
    grant_script: Optional[str] = None


class RmOracleUserAddMeta(RmMetaBase):
    allow_login: bool = True
    allow_resource: bool = True
    roles: List[str] = []


class RmOracleRoleAddMeta(RmMetaBase):
    not_identified: bool = False,
    identified_by_password: Optional[str] = None
    identified_using: Optional[str] = None
    identified_externally: bool = False
    identified_globally: bool = False
    identified_globally_as: Optional[str] = None
    container: Optional[str] = None

    grant_script: Optional[str] = None


class RmMongoDbUserAddMeta(RmMetaBase):
    roles: List[str] = []


class RmMongoDbRoleAddMeta(RmMetaBase):
    js_script: Optional[str] = None


# MACHINE


class RmLinuxGroupAddMeta(RmMetaBase):
    gid: Optional[int] = None
    system_group: Optional[bool] = False


class RmMachineUserAddMeta(RmMetaBase):
    use_password: Optional[bool] = True
    use_private_key: Optional[bool] = False
    use_private_key_passphrase: Optional[bool] = False
    use_private_key_type: Optional[str] = "ecdsa_sha2_nistp521"
    private_key: Optional[str] = None
    private_key_passphrase: Optional[str] = None
    authorized_keys: List[str] = []


class RmLinuxUserAddMeta(RmMachineUserAddMeta):
    system_user: Optional[bool] = False
    shell: Optional[str] = None
    no_login:  Optional[bool] = False
    home_dir: Optional[str] = None
    do_not_create_home_dir: Optional[bool] = False
    allow_bad_names: Optional[bool] = False
    gecos_full_name: Optional[str] = None
    gecos_room_number: Optional[str] = None
    gecos_work_phone: Optional[str] = None
    gecos_home_phone: Optional[str] = None
    gecos_other: Optional[str] = None
    group: Optional[str] = None
    groups: List[str] = []
    create_group: Optional[bool] = False
    validate_group: Optional[bool] = True
    uid: Optional[str] = None
    selinux_user_context:  Optional[str] = None
    btrfs_subvolume: Optional[bool] = False
    system_dir_mode: Optional[str] = None
    non_system_dir_mode: Optional[str] = None


class RmLinuxUserDeleteMeta(RmMetaBase):
    remove_home_dir: Optional[bool] = True
    remove_user_group: Optional[bool] = True


class RmWindowsGroupAddMeta(RmMetaBase):
    description: Optional[str] = None


class RmWindowsUserAddMeta(RmMachineUserAddMeta):
    display_name: Optional[str] = None
    description: Optional[str] = None
    disabled: bool = False
    expire_days: int = 0
    groups: List[str] = []


class RmMacOsUserAddMeta(RmMachineUserAddMeta):
    display_name: Optional[str] = None
    uid: Optional[str] = None
    gid: Optional[str] = None
    shell: Optional[str] = None
    home_dir: Optional[str] = None
    is_admin: bool = False
    is_role_account: bool = False
    groups: List[str] = []


class RmMacOsRoleAddMeta(RmMetaBase):
    display_name: Optional[str] = None
    gid: Optional[str] = None
    record_name: Optional[str] = None


# DIRECTORY


class RmBaseLdapUserAddMeta(RmMetaBase):
    object_class: List[str] = []
    dn: Optional[str] = None
    base_dn: Optional[str] = None
    auto_uid_number: bool = True
    gid_number_match_uid: bool = True
    home_dir_base: Optional[str] = "/home"
    first_rdn_component: Optional[str] = None
    attributes: Optional[dict] = {}
    groups: List[str] = []


class RmOpenLdapUserAddMeta(RmBaseLdapUserAddMeta):
    """
    Parameters for creating a user for OpenLDAP

    :type: directory,openldap,create,user
    :param object_class: Object classes for the entry.
    :param dn: Full distinguished name for the user.
    :param base_dn: Base part of the distinguished name.
    :param auto_uid_number: Automatically set the UID for the user.
    :param gid_number_match_uid: Set the GID to the UID.
    :param home_dir_base: Absolute home directory path.
    :param first_rdn_component: The Relative Distinguished Name component.
    :param attributes: Key/value pairs of attributes.
    :param groups: Add the user to the follow group DN entries.
    """

    object_class: List[str] = ["top", "inetOrgPerson", "posixAccount"]


class RmAdUserAddMeta(RmBaseLdapUserAddMeta):
    """
    Parameters for creating a user for Active Directory

    :type: directory,active_directory,create,user
    :param object_class: Object classes for the entry.
    :param dn: Full distinguished name for the user.
    :param base_dn: Base part of the distinguished name.
    :param auto_uid_number: Automatically set the UID for the user.
    :param gid_number_match_uid: Set the GID to the UID.
    :param home_dir_base: Absolute home directory path.
    :param first_rdn_component: The Relative Distinguished Name component.
    :param attributes: Key/value pairs of attributes.
    :param groups: Add the user to the follow group DN entries.
    :param user_account_control: User account control flag.
    """

    object_class: List[str] = ["top", "person", "organizationalPerson", "user"]
    user_account_control: Optional[int] = 512


class RmBaseLdapGroupAddMeta(RmMetaBase):
    object_class: List[str] = []
    dn: Optional[str] = None
    ou_group: Optional[str] = None
    description: Optional[str] = None

    # Group should use CN, users can use UID, but allow it to be changed; default to CN
    first_rdn_component: str = "CN"

    attributes: Optional[dict] = {}


class RmOpenLdapGroupAddMeta(RmBaseLdapGroupAddMeta):
    object_class: List[str] = ["top", "groupOfNames"]
    ou_group: Optional[str] = "OU=groups"


class RmAdGroupAddMeta(RmBaseLdapGroupAddMeta):
    object_class: List[str] = ["top", "group"]
    ou_group: Optional[str] = "CN=Users"

    # These are fake LDAP attributes.
    # They are using to calculate a group_type.
    group_category: str = "Security"
    group_scope: str = "Global"

    # If this 0, a value will be calculated from the category and scrope.
    # If this is now 0, then his value will be used.
    group_type: int = 0


class RmBaseLdapUserDeleteMeta(RmMetaBase):
    orphan_check: Optional[bool] = False
    delete_orphans: Optional[bool] = True


class RmOpenLdapUserDeleteMeta(RmBaseLdapUserDeleteMeta):
    pass


class RmAdUserDeleteMeta(RmBaseLdapUserDeleteMeta):
    pass



#  _  __
# | |/ /___ ___ _ __  ___ _ _ ®
# | ' </ -_) -_) '_ \/ -_) '_|
# |_|\_\___\___| .__/\___|_|
#              |_|
#
# Keeper Commander
# Copyright 2024 Keeper Security Inc.
# Contact: ops@keepersecurity.com
#

import json
import logging
from pathlib import Path
import shutil
from typing import Dict, Any, Optional
from configparser import ConfigParser

import yaml
from .file_handler import ConfigFormatHandler
from .models import ServiceConfigData
from ..decorators.logging import logger, debug_decorator
from ..util.exceptions import ValidationError
from ... import resources, utils
from ...params import KeeperParams


VALID_CERT_EXTENSIONS = {".pem", ".crt", ".cer", ".key"}
class ServiceConfig:
    def __init__(self, title: str = 'Commander Service Mode'):
        self.title = title
        
        self.config = ConfigParser()

        config_ini_path = Path(resources.__file__).parent / 'service_config.ini'
        self.config.read(config_ini_path)
        self.messages = self.config['Messages']
        self.validation_messages = self.config['Validation_Messages']

        default_path = utils.get_default_path()

        self.format_handler = ConfigFormatHandler(
            config_dir=default_path,
            messages=self.messages,
            validation_messages=self.validation_messages
        )
        self.config_path = self.format_handler.config_path

        self._validator = None
        self._cli_handler = None
        self._command_validator = None
        self._record_handler = None

    @property
    def validator(self):
        if self._validator is None:
            from .config_validation import ConfigValidator
            self._validator = ConfigValidator()
        return self._validator

    @property
    def cli_handler(self):
        if self._cli_handler is None:
            from .cli_handler import CommandHandler
            self._cli_handler = CommandHandler()
        return self._cli_handler

    @property
    def command_validator(self):
        if self._command_validator is None:
            from .command_validator import CommandValidator
            self._command_validator = CommandValidator()
        return self._command_validator

    @property
    def record_handler(self):
        if self._record_handler is None:
            from .record_handler import RecordHandler
            self._record_handler = RecordHandler()
        return self._record_handler

    @debug_decorator
    def create_default_config(self) -> Dict[str, Any]:
        """Create default configuration structure."""
        config = ServiceConfigData(
            title=self.title,
            port=None,
            ngrok="n",
            ngrok_auth_token="",
            ngrok_custom_domain="",
            ngrok_public_url="",
            cloudflare="n",
            cloudflare_tunnel_token="",
            cloudflare_custom_domain="",
            cloudflare_public_url="",
            tls_certificate="n",
            certfile="",
            certpassword="",
            is_advanced_security_enabled="n",
            rate_limiting="",
            ip_allowed_list="",
            ip_denied_list="",
            encryption="",
            encryption_private_key="",
            fileformat="yaml",
            run_mode="foreground",
            queue_enabled="y",
            records=[]
        ).__dict__
        return config

    def save_config(self, config_data: Dict[str, Any], save_type: str = None) -> Path:
        """Save configuration to file."""
        self._validate_config_structure(config_data)
        return self.format_handler.save_config(config_data, save_type)

    

    def get_cert_paths(self, config_data: Dict[str, Any]) -> Dict[str, Path]:
        """Retrieve certificate file paths with validation."""
        cert_paths = {}
        for key in ["certfile", "certpassword"]:
            if config_data.get(key):
                file_path = Path(config_data[key])
                if file_path.suffix in VALID_CERT_EXTENSIONS:
                    cert_paths[key] = file_path
                else:
                    raise ValueError(f"Invalid file format for {key}: {file_path}. Allowed: {VALID_CERT_EXTENSIONS}")
        return cert_paths

    def update_service_config(self, updates: Dict[str, str]) -> None:
        """
        Update specified keys in the service configuration file (.json or .yml).

        Args:
            updates: Dictionary where each key is a config field (like 'certfile')
                    and value is the **certificate file name** stored in `.keeper`.
        """
        try:
            config_dir = utils.get_default_path()
            config_dir.mkdir(parents=True, exist_ok=True)  # Ensure directory exists

            # Find any service_config file
            config_path = None
            file_format = None

            for file in config_dir.glob("service_config.*"):
                ext = file.suffix.lower()
                if ext in [".json", ".yml", ".yaml"]:
                    config_path = file
                    file_format = "json" if ext == ".json" else "yaml"
                    break

            if config_path: 
                self.read_plain_text_config_file(config_path, file_format, updates)
            else:
                logging.info("No existing service_config file found.")

        except Exception as e:
            logging.info(f"Error updating service configuration: {e}")
    
    # Read plain text config file and update configuration
    def read_plain_text_config_file(self, config_path: str, file_format: str, updates: Dict[str, str]) -> None:
        """ Read plain text config file and update configuration """
        with open(config_path, "r") as f:
                     config_data = json.load(f) if file_format == "json" else yaml.safe_load(f) or {}
        # Save updated configuration
        with open(config_path, "w") as f:
            if file_format == "json":
                    json.dump(config_data, f, indent=4)
            else:
                yaml.safe_dump(config_data, f, default_flow_style=False)
 
        logging.info(f"Updated keys in {config_path.name}: {', '.join(updates.keys())}")

    def save_cert_data(self, config_data: Dict[str, Any], save_type: str = None) -> Path:
        """Save certificate and password files in the .keeper folder and update service_config.json/.yml."""
        try:
            keeper_dir = utils.get_default_path()
            keeper_dir.mkdir(parents=True, exist_ok=True)

            cert_paths = self.get_cert_paths(config_data)
            
            if not cert_paths:
                return keeper_dir

            updated_names = {}
            saved_files = []

            for key, src_path in cert_paths.items():
                dest_path = keeper_dir / src_path.name
                if src_path.exists():
                    shutil.copy(src_path, dest_path)
                    saved_files.append(dest_path)
                    updated_names[key] = src_path.name  # Store only the filename, not the full path
                else:
                    raise FileNotFoundError(f"File not found: {src_path}")

            self.update_service_config(updated_names)

            if saved_files:
                logging.info(f"Certificates saved in {keeper_dir}: {', '.join(str(f.name) for f in saved_files)}")
            return keeper_dir

        except Exception as e:
            logging.info(f"Error saving certificate data: {e}")
            return None

    def load_config(self) -> Dict[str, Any]:
        """Load configuration from file."""
        config = self.format_handler.load_config()
        
        # Add backwards compatibility for missing queue_enabled field
        if 'queue_enabled' not in config:
            config['queue_enabled'] = 'y'  # Default to enabled for existing configs
            logger.debug("Added default queue_enabled=y for backwards compatibility")

        # Add backwards compatibility for missing Cloudflare fields
        if 'cloudflare' not in config:
            config['cloudflare'] = 'n'  # Default to disabled for existing configs
            logger.debug("Added default cloudflare=n for backwards compatibility")
        
        if 'cloudflare_tunnel_token' not in config:
            config['cloudflare_tunnel_token'] = ''
            logger.debug("Added default cloudflare_tunnel_token for backwards compatibility")
        
        if 'cloudflare_custom_domain' not in config:
            config['cloudflare_custom_domain'] = ''
            logger.debug("Added default cloudflare_custom_domain for backwards compatibility")
        
        if 'cloudflare_public_url' not in config:
            config['cloudflare_public_url'] = ''
            logger.debug("Added default cloudflare_public_url for backwards compatibility")
        
        self._validate_config_structure(config)
        return config

    @debug_decorator
    def _validate_config_structure(self, config: Dict[str, Any]) -> None:
        """Validate the configuration structure."""
        try:
            config_data = ServiceConfigData(**config)
        except TypeError as e:
            raise ValidationError(f"Invalid configuration structure: {str(e)}")

        if config_data.ngrok == 'y':
            logger.debug("Validating ngrok configuration")
            self.validator.validate_ngrok_token(config_data.ngrok_auth_token)

        if config_data.cloudflare == 'y':
            logger.debug("Validating cloudflare configuration")
            self.validator.validate_cloudflare_token(config_data.cloudflare_tunnel_token)
            self.validator.validate_domain(config_data.cloudflare_custom_domain)

        if config_data.is_advanced_security_enabled == 'y':
            logger.debug("Validating advanced security settings")
            self.validator.validate_rate_limit(config_data.rate_limiting)
            self.validator.validate_ip_list(config_data.ip_allowed_list)
            self.validator.validate_ip_list(config_data.ip_denied_list)

        if config_data.encryption == 'y':
            logger.debug("Validating encryption settings")
            self.validator.validate_encryption_key(config_data.encryption_private_key)

    def _get_yes_no_input(self, prompt: str) -> str:
        logger.debug(f"Requesting y/n input with prompt: {prompt}")
        while True:
            if (user_input := input(prompt).lower()) in ['y', 'n']:
                logger.debug(f"Received valid input: {user_input}")
                return user_input
            print(self.validation_messages['invalid_input'])

    def validate_command_list(self, commands: str, params: KeeperParams = None) -> str:
        """Validate command list against available Keeper Commander commands."""
        if not commands:
            raise ValidationError("Command list cannot be empty")
        help_output = self.cli_handler.get_help_output(params)
        valid_commands, command_info = self.command_validator.parse_help_output(help_output)
        validated_commands, invalid_commands = self.command_validator.validate_command_list(commands, valid_commands)
        
        if invalid_commands:
            error_msg = self.command_validator.generate_command_error_message(invalid_commands, command_info)
            raise ValidationError(error_msg)

        return validated_commands

    def _get_validated_commands(self, params: KeeperParams) -> str:
        """Get and validate command list from user input."""
        while True:
            try:
                command_list = input(self.messages['command_list_prompt'])
                return self.validate_command_list(command_list, params)
            except ValidationError as e:
                print(f"\nError: {str(e)}")
                print("\nPlease try again with valid commands.")

    def create_record(self, is_advanced_security_enabled: str, params: KeeperParams, commands: Optional[str] = None) -> Dict[str, Any]:
        """Create a new configuration record."""
        commands = self.validate_command_list(commands, params) if commands else self._get_validated_commands(params)
        return self.record_handler.create_record(is_advanced_security_enabled, commands)

    def update_or_add_record(self, params: KeeperParams) -> None:
        """Update existing record or add new one."""
        self.record_handler.update_or_add_record(params, self.title, self.format_handler.config_path)
        logger.debug(f" Uploaded file remote success.")
        self.format_handler.encrypt_config_file(self.format_handler.config_path, self.format_handler.config_dir)
        logger.debug(f" Local file encryption success.")
        try:
            config = self.load_config()
            if config.get("tls_certificate") == "y":
                self.record_handler.update_or_add_cert_record(params, self.title)
                logger.debug(f" Uploaded TLS certificate at remote.")
            else:
                logger.debug(f" TLS certificate upload skipped - TLS not enabled.")
        except Exception as e:
            logger.debug(f"Error checking TLS configuration: {e}")